"""SELECT statement builder."""

from sqlfactory.mixins.join import CrossJoin, InnerJoin, Join, LeftJoin, LeftOuterJoin, RightJoin, RightOuterJoin
from sqlfactory.select.aliased import Aliased, SelectColumn
from sqlfactory.select.column_list import ColumnList
from sqlfactory.select.select import SELECT, Select

__all__ = [
    "SELECT",
    "Aliased",
    "ColumnList",
    "CrossJoin",
    "InnerJoin",
    "Join",
    "LeftJoin",
    "LeftOuterJoin",
    "RightJoin",
    "RightOuterJoin",
    "Select",
    "SelectColumn",
]



"""Simple binary comparison conditions."""

from typing import Any, TypeAlias

from sqlfactory.condition.base import Condition, StatementOrColumn
from sqlfactory.statement import Statement


class SimpleCondition(Condition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Simple condition comparing one column with given value, using specified operator.
    """

    def __init__(self, column: StatementOrColumn, operator: str, value: Statement | Any) -> None:
        # pylint: disable=duplicate-code   # It does not make sense to generalize two-row statement used on two places.
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param operator: Operator to use for comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if not isinstance(column, Statement):
            # pylint: disable=import-outside-toplevel,cyclic-import
            from sqlfactory.entities import Column

            column = Column(column)

        args = []

        if isinstance(column, Statement):
            args.extend(column.args)

        if isinstance(value, Statement):
            args.extend(value.args)

        elif not isinstance(value, Statement):
            args.append(value)

        if isinstance(value, Statement):
            super().__init__(f"{column!s} {operator} {value!s}", *args)
        else:
            super().__init__(f"{column!s} {operator} %s", *args)


class Equals(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Equality condition (`==`). You can also use shorthand alias `Eq`.

    Note that first argument to `Equals` (or `Eq`) is expected to be column, while second argument is value. So to compare
    two columns, you must use `Column` instances as second argument (`Eq("column1", Column("column2"))` to produce
    ``` `column1` = `column2` ```).

    ```Eq("column1", "column2")``` produces ``` `column1` = %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c == "value"  # == Eq("column", "value")
    c == None     # == Eq("column", None)
    c == Now()    # == Eq("column", Now())
    ```

    Examples:

    - Simple comparison of column to value
        ```python
        # `column` = <value>
        Eq("column", "value")
        "`column` = %s", ["value"]
        ```
    - Comparison column to None
        ```python
        # `column` IS NULL
        Eq("column", None)
        "`column` IS NULL"
        ```
    - Comparison of generic statement to value
        ```python
        # <statement> = <value>
        Eq(Date(), "2021-01-01")
        "DATE() = %s", ["2021-01-01"]
        ```
    - Comparison of generic statement to None
        ```python
        # <statement> IS NULL
        Eq(Date(), None)
        "DATE() IS NULL"
        ```
    - Comparison of statement to statement
        ```python
        # <statement> = <statement>
        Eq(Date(), Now())
        # "DATE() = NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | None | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS", value)
        else:
            super().__init__(column, "=", value)


class NotEquals(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Not equality condition (`!=`). You can also use shorthand alis `Ne`.

    Note that first argument to `NotEquals` (or `Ne`) is expected to be column, while second argument is value. So to compare
    two columns, you must use `Column` instances as second argument (`Ne("column1", Column("column2"))` to produce
    ``` `column1` != `column2` ```).

    ```Ne("column1", "column2")``` produces ``` `column1` != %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c != "value"  # == Ne("column", "value")
    c != None     # == Ne("column", None)
    c != Now()    # == Ne("column", Now())
    ```

    Examples:

    - Column not equals value
        ```python
        # `column` != <value>
        Ne("column", "value")
        "`column` != %s", ["value"]
        ```
    - Statement not equals value
        ```python
        # <statement> != <value>
        Ne(Date(), "2021-01-01")
        "DATE() != %s", ["2021-01-01"]
        ```
    - Column is not None
        ```python
        # `column` IS NOT NULL
        Ne("column", None)
        "`column` IS NOT NULL"
        ```
    - Statement is not None
        ```python
        # <statement> IS NOT NULL
        Ne(Date(), None)
        "DATE() IS NOT NULL"
        ```
    - Statement not equals statement
        ```python
        # <statement> != <statement>
        Ne(Date(), Now())
        "DATE() != NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | None | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS NOT", value)
        else:
            super().__init__(column, "!=", value)


class GreaterThanOrEquals(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Greater than or equal condition (`>=`). You can also use shorthand alis `Ge`.

    Note that first argument to `GreaterThanOrEquals` (or `Ge`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Ge("column1", Column("column2"))` to produce
    ``` `column1` >= `column2` ```).

    ```Ge("column1", "column2")``` produces ``` `column1` >= %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c >= "value"  # == Ge("column", "value")
    c >= None     # == Ge("column", None)
    c >= Now()    # == Ge("column", Now())
    ```

    Examples:

    - Column is greater or equals to value:
        ```python
        # `column` >= <value>
        Ge("column", 10)
        "`column` >= %s", [10]
        ```
    - Statement is greater or equals to value:
        ```python
        # <statement> >= <value>
        Ge(Date(), "2021-01-01")
        "DATE() >= %s", ["2021-01-01"]
        ```
    - Column is greater or equals to other column
        ```python
        # `column1` >= `column2`
        Ge("column1", Column("column2"))
        "`column1` >= `column2`"
        ```
    - Column is greater or equals to statement
        ```python
        # `column` >= <statement>
        Ge("column", Now())
        "`column` >= NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">=", value)


class GreaterThan(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Greater than condition (`>`). You can also use shorthand alis `Gt`.

    Note that first argument to `GreaterThan` (or `Ge`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Gt("column1", Column("column2"))` to produce
    ``` `column1` > `column2` ```).

    ```Gt("column1", "column2")``` produces ``` `column1` > %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c > "value"  # == Gt("column", "value")
    c > None     # == Gt("column", None)
    c > Now()    # == Gt("column", Now())
    ```

    Examples:

    - Column is greater than value:
        ```python
        # `column` > <value>
        Gt("column", 10)
        "`column` > %s, [10]
        ```
    - Statement is greater than value:
        ```python
        # <statement> > <value>
        Gt(Date(), "2021-01-01")
        "DATE() > %s", ["2021-01-01"]
        ```
    - Column is greater than other column
        ```python
        # `column1` > `column2`
        Gt("column1", Column("column2"))
        "`column1` > `column2`"
        ```
    - Column is greater than statement
        ```python
        # `column` > <statement>
        Gt("column", Now())
        "`column` > NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">", value)


class LessThanOrEquals(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Less than or equal condition (`<=`). You can also use shorthand alis `Le`.

    Note that first argument to `LessThanOrEquals` (or `Le`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Le("column1", Column("column2"))` to produce
    ``` `column1` <= `column2` ```).

    ```Le("column1", "column2")``` produces ``` `column1` <= %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c <= "value"  # == Le("column", "value")
    c <= None     # == Le("column", None)
    c <= Now()    # == Le("column", Now())
    ```

    Examples:

    - Column is lower or equal to value:
        ```python
        # `column` <= <value>
        Le("column", 10)
        "`column` <= %s", [10]
        ```
    - Statement is lower or equal to value
        ```python
        <statement> <= <value>
        Le(Date(), "2021-01-01")
        "DATE() <= %s", ["2021-01-01"]
        ```
    - Column is lower or equal to other column
        ```python
        # `column1` <= `column2`
        Le("column1", Column("column2"))
        "`column1` <= `column2`"
        ```
    - Column is lower or equal to statement
        ```python
        # `column` <= <statement>
        Le("column", Now())
        "`column` <= NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<=", value)


class LessThan(SimpleCondition):
    # pylint: disable=too-few-public-methods  # As everything is handled in base classes.
    """
    Less than condition (`<`). You can also use shorthand alis `Lt`.

    Note that first argument to `LessThan` (or `Lt`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Lt("column1", Column("column2"))` to produce
    ``` `column1` < `column2` ```).

    ```Lt("column1", "column2")``` produces ``` `column1` < %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c < "value"  # == Lt("column", "value")
    c < None     # == Lt("column", None)
    c < Now()    # == Lt("column", Now())
    ```

    Examples:

    - Column is lower than value:
        ```python
        # `column` < <value>
        Lt("column", 10)
        "`column` < %s", [10]
        ```
    - Statement is lower than value:
        ```python
        # <statement> < <value>
        Lt(Date(), "2021-01-01")
        "DATE() < %s", ["2021-01-01"]
        ```
    - Column is lower than other column
        ```python
        # `column1` < `column2`
        Lt("column1", Column("column2"))
        "`column1` < `column2`"
        ```
    - Column is lower than statement
        ```python
        # `column` < <statement>
        Lt("column", Now())
        "`column` < NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<", value)


# Convenient aliases for shorter code.
Eq: TypeAlias = Equals
Ge: TypeAlias = GreaterThanOrEquals
Gt: TypeAlias = GreaterThan
Le: TypeAlias = LessThanOrEquals
Lt: TypeAlias = LessThan
Ne: TypeAlias = NotEquals



"""WHERE mixin for query generator."""

from __future__ import annotations

from typing import Any, Generic, Self, TypeVar

from sqlfactory.condition.base import And, ConditionBase

T = TypeVar("T")


class WithWhere(Generic[T]):
    """
    Mixin to provide `WHERE` support for query generator.

    Example:

    >>> from sqlfactory import Select, And, Equals
    >>> Select(where=And(Equals("id", 1), Equals("status", "active")))

    >>> from sqlfactory import Update, And, Equals
    >>> Update(where=And(Equals("id", 1), Equals("status", "active")))

    >>> from sqlfactory import Delete, And, Equals
    >>> Delete(where=And(Equals("id", 1), Equals("status", "active")))

    """

    def __init__(self, *args: Any, where: ConditionBase | None = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._where = where

    def where(self, condition: ConditionBase) -> Self:
        """
        Set `WHERE` condition for query. If you chain multiple where calls together, they are joined by `And` condition.
        You can manually specify joining condition by use one of `sqlfactory.And` or `sqlfactory.Or` compound classes to chain
        multiple conditions.

        Example:

        >>> from sqlfactory import Select, And, Equals
        >>> sel = Select().where(And(Equals("id", 1), Equals("status", "active")))

        :param condition: Condition to be used in WHERE clause.
        """
        if self._where is not None:
            if isinstance(self._where, And):
                self._where.append(condition)
            else:
                self._where = And(self._where, condition)
        else:
            self._where = condition

        return self

    # pylint: disable=invalid-name
    def WHERE(self, condition: ConditionBase) -> Self:
        """Alias for `WithWhere.where()` to be more SQL-like with all capitals."""
        return self.where(condition)



"""LIKE statement"""

from typing import Any

from sqlfactory.condition.base import Condition, StatementOrColumn
from sqlfactory.entities import Column
from sqlfactory.statement import Statement


class Like(Condition):
    """
    SQL `LIKE` condition for comparing strings against pattern.

    Examples:

    - Simple
        ```python
        # `column` LIKE %s
        Like("column", "pattern")
        "`column` LIKE %s", ["pattern"]
        ```
    - Negative
        ```python
        # `column` NOT LIKE %s
        Like("column", "pattern", negative=True)
        "`column` NOT LIKE %s", ["pattern"]
        ```
    - Statement
        ```python
        Like("column", Concat("%", Column("other_column"), "%"))
        "`column` LIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Statement (negative)
        ```python
        Like("column", Concat("%", Column("other_column"), "%"), negative=True)
        "`column` NOT LIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Instead of column, you can also use any other expression
        ```python
        Like(Concat("column", "other_column"), "pattern")
        "CONCAT(`column`, `other_column`) LIKE %s", ["pattern"]
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement, negative: bool = False) -> None:
        """
        :param column: Column (or statement) on left side of LIKE operator.
        :param value: Value to match the column (or statement) against.
        :param negative: Whether to use negative matching (NOT LIKE).
        """
        args = []

        if not isinstance(column, Statement):
            column = Column(column)

        if isinstance(column, Statement):
            args.extend(column.args)

        if isinstance(value, Statement):
            args.extend(value.args)
        else:
            args.append(value)

        if isinstance(value, Statement):
            super().__init__(
                f"{column!s}{' NOT' if negative else ''} LIKE {value!s}",
                *args,
            )
        else:
            super().__init__(f"{column!s}{' NOT' if negative else ''} LIKE %s", *args)

    @staticmethod
    def escape(s: str) -> str:
        """
        Escape string for use in LIKE statement
        :param s: String to be escaped
        :return: String with escaped characters - % -> %%, _ -> __, to be safely used as part of pattern in LIKE statement.
        """
        return s.replace("%", "%%").replace("_", "__")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Collection, Generic, Self, TypeVar, overload

from sqlfactory.condition import ConditionBase
from sqlfactory.entities import Table
from sqlfactory.statement import Statement

if TYPE_CHECKING:
    from sqlfactory.select import Select  # pragma: no cover

T = TypeVar("T")


class Join(Statement):
    """
    Produces `JOIN` statement

    Usage:

    >>> from sqlfactory import Join, Eq
    >>> Join("table2", on=Eq("table1.id", "table2.id"))
    >>> "JOIN `table2` ON `table1`.`id` = `table2`.`id`"

    >>> from sqlfactory import Join, Eq
    >>> Join("table2", on=Eq("table1.id", "t2.id"), alias="t2")
    >>> "JOIN `table2` AS `t2` ON `table1`.`id` = `t2`.`id`"

    With subquery:

    >>> from sqlfactory import Join, Eq, Select, Column
    >>> from sqlfactory.func.agg import Sum
    >>> Join(
    ...     Select("table2.product_id", Sum("price"), table="table2", where=Eq("table2.enabled", True)),
    ...     alias="t2",
    ...     on=Eq("t2.product_id", Column("table.id"))
    ... )
    >>> "JOIN ("
    ... "SELECT `table2`.`product_id`, SUM(`price`) "
    ... "FROM `table2` "
    ... "WHERE `table2`.`enabled` = %s"
    ... ") AS `t2` ON `t2`.`product_id` = `table`.`id`"

    Note that for subquery join, alias must be always specified.
    """

    def __init__(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> None:
        """
        :param table: Table to be joined
        :param on: ON condition
        :param alias: Optional alias of the joined table.
        """
        if isinstance(table, str):
            table = Table(table)

        self.table = table
        """Table to be joined"""

        self.on = on
        """ON join condition"""

        self.alias = alias
        """Optional alias for the joined table."""

        from sqlfactory.select import Select  # pylint: disable=import-outside-toplevel, cyclic-import

        if isinstance(self.table, Select) and not self.alias:
            raise AttributeError("When joining a subselect, alias must be specified.")

    @property
    def join_spec(self) -> str:
        """
        Returns the JOIN type itself for generation of SQL query.
        @private
        """
        return "JOIN"

    def __str__(self) -> str:
        from sqlfactory.select import Select  # pylint: disable=import-outside-toplevel, cyclic-import

        if self.alias:
            if isinstance(self.table, Select):
                table = f"({self.table!s}) AS `{self.alias}`"
            else:
                table = f"{self.table!s} AS `{self.alias}`"
        else:
            table = str(self.table)

        if self.on:
            return f"{self.join_spec} {table} ON {self.on!s}"

        return f"{self.join_spec} {table}"

    @property
    def args(self) -> list[Any]:
        """Argument values of the JOIN statement."""
        return [*(self.table.args if isinstance(self.table, Statement) else []), *(self.on.args if self.on else [])]


class LeftJoin(Join):
    """
    Produces `LEFT JOIN` statement

    Usage:

    >>> from sqlfactory import LeftJoin, Eq, Column
    >>> LeftJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "LEFT JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "LEFT JOIN"


class LeftOuterJoin(Join):
    """
    Produces `LEFT OUTER JOIN` statement

    Usage:

    >>> from sqlfactory import LeftOuterJoin, Eq, Column
    >>> LeftOuterJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "LEFT OUTER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "LEFT OUTER JOIN"


class RightJoin(Join):
    """
    Produces `RIGHT JOIN` statement

    Usage:

    >>> from sqlfactory import RightJoin, Eq, Column
    >>> RightJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "RIGHT JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "RIGHT JOIN"


class RightOuterJoin(Join):
    """
    Produces `RIGHT OUTER JOIN` statement

    Usage:

    >>> from sqlfactory import RightOuterJoin, Eq, Column
    >>> RightOuterJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "RIGHT OUTER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "RIGHT OUTER JOIN"


class InnerJoin(Join):
    """
    Produces `INNER JOIN` statement

    Usage:

    >>> from sqlfactory import InnerJoin, Eq, Column
    >>> InnerJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "INNER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "INNER JOIN"


class CrossJoin(Join):
    """
    Produces `CROSS JOIN` statement

    Usage:

    >>> from sqlfactory import CrossJoin, Eq, Column
    >>> CrossJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "CROSS JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "CROSS JOIN"


class WithJoin(Generic[T]):
    """Mixin to provide JOIN support for query generator."""

    def __init__(self, *args: Any, join: Collection[Join] | None = None, **kwargs: Any) -> None:
        """
        :param join: List of JOIN clauses
        """
        super().__init__(*args, **kwargs)
        self._join = list(join) if join is not None else []

    def _append_join(self, join: Join) -> Self:
        """Append join to list of joins."""
        if not self._join:
            self._join = []

        if join not in self._join:
            self._join.append(join)

        return self

    @overload
    def join(self, join: Join, /) -> Self:
        """Append JOIN clause to the query (any Join instance)."""

    @overload
    def join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>)"""

    def join(self, table: str | Table | Join | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>)
        """
        if isinstance(table, Join):
            if on is not None or alias is not None:
                raise AttributeError("When passing Join instance directly, on or alias attributes cannot be specified.")

            return self._append_join(table)

        return self._append_join(Join(table, on, alias))

    @overload
    def JOIN(self, join: Join, /) -> Self:  # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""

    @overload
    def JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""

    def JOIN(self, table: str | Table | Select | Join, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""
        return self.join(table, on, alias)  # type: ignore[arg-type]  # mypy searches in overloads

    def left_join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """Append LEFT JOIN clause to the query."""
        return self.join(LeftJoin(table, on, alias))

    def LEFT_JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for left_join() to be more SQL-like with all capitals."""
        return self.left_join(table, on, alias)

    def right_join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """Append RIGHT JOIN clause to the query."""
        return self.join(RightJoin(table, on, alias))

    def RIGHT_JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for right_join() to be more SQL-like with all capitals."""
        return self.right_join(table, on, alias)



"""BETWEEN condition generator"""

from typing import Any

from sqlfactory.condition.base import Condition, StatementOrColumn
from sqlfactory.entities import Column
from sqlfactory.statement import Statement


# pylint: disable=too-few-public-methods  # As everything is handled by base classes.
class Between(Condition):
    # pylint: disable=duplicate-code  # It does not make sense to generalize two-row statement used on two places.
    """
    Provides generation for following syntax:

    - ``` `column` BETWEEN <lower_bound> AND <upper_bound>```
    - ``` `column` NOT BETWEEN <lower_bound> AND <upper_bound>```
    - ```<statement> BETWEEN <lower_bound> AND <upper_bound>```
    - ```<statement> NOT BETWEEN <lower_bound> AND <upper_bound>```

    Usage:

    >>> Between('column', 1, 10)
    >>> "`column` BETWEEN 1 AND 10"

    >>> Between('column', 1, 10, negative=True)
    >>> "`column` NOT BETWEEN 1 AND 10"

    >>> Between(Column('c1') + Column('c2'), 1, 10)
    >>> "(`c1` + `c2`) BETWEEN 1 AND 10"

    >>> Between(Column('c1') + Column('c2'), Column('c3') + Column('c4'), Column('c5') + Column('c6'))
    >>> "(`c1` + `c2`) BETWEEN (`c3` + `c4`) AND (`c5` + `c6`)"

    """

    def __init__(
        self, column: StatementOrColumn, lower_bound: Any | Statement, upper_bound: Any | Statement, *, negative: bool = False
    ) -> None:
        """
        :param column: Column to be compared.
        :param lower_bound: Lower inclusive bound of matching value
        :param upper_bound: Upper inclusive bound of matching value
        :param negative: Whether to negate the condition.
        """

        lower_bound_s = "%s"
        upper_bound_s = "%s"

        if not isinstance(column, Statement):
            column = Column(column)

        args = []

        if isinstance(column, Statement):
            args.extend(column.args)

        if isinstance(lower_bound, Statement):
            lower_bound_s = str(lower_bound)
            if isinstance(lower_bound, Statement):
                args.extend(lower_bound.args)
        else:
            args.append(lower_bound)

        if isinstance(upper_bound, Statement):
            upper_bound_s = str(upper_bound)
            if isinstance(upper_bound, Statement):
                args.extend(upper_bound.args)
        else:
            args.append(upper_bound)

        super().__init__(f"{column!s} {'NOT ' if negative else ''}BETWEEN {lower_bound_s} AND {upper_bound_s}", *args)



"""ORDER BY mixin for query generator"""

from __future__ import annotations

from enum import Enum
from typing import Any, Collection, Generic, Literal, Optional, Self, TypeVar

from sqlfactory.entities import Column, ColumnArg
from sqlfactory.statement import Statement


class Direction(str, Enum):
    """
    Ordering direction as enum

    Usage:

    >>> from sqlfactory import Order, Direction
    >>> Order([('column1', Direction.ASC), ('column2', Direction.DESC)])
    >>> "ORDER BY `column1` ASC, `column2` DESC"
    """

    ASC = "ASC"
    DESC = "DESC"


OrderColumn = ColumnArg | Statement


class Order(list[tuple[OrderColumn, Direction | Literal["ASC", "DESC"]]], Statement):  # type: ignore[misc]
    """
    ORDER BY statement as list of columns to use for ordering. For usage with SELECT, UPDATE and DELETE statements.

    Usage:

    >>> from sqlfactory import Order, Direction
    >>> Order([('column1', Direction.ASC), ('column2', Direction.DESC)])
    >>> "ORDER BY `column1` ASC, `column2` DESC"
    """

    def __str__(self) -> str:
        if not self:
            return ""

        out = []

        for column, direction in self:
            if isinstance(column, str):
                column = Column(column)

            out.append(f"{column!s} {direction.value if isinstance(direction, Direction) else direction}")

        return f"ORDER BY {', '.join(out)}"

    @property
    def args(self) -> list[Any]:
        """Argument values for the order by statement"""
        out = []
        for column, _ in self:
            if isinstance(column, Statement):
                out.extend(column.args)

        return out


T = TypeVar("T")


class WithOrder(Generic[T]):
    """Mixin to provide ORDER BY support for query generator."""

    def __init__(self, *args: Any, order: OrderArg | None = None, **kwargs: Any) -> None:
        """

        Example:

        >>> from sqlfactory import Select, Direction
        >>> query = Select(order=[('column1', Direction.ASC), ('column2', Direction.DESC)])

        >>> from sqlfactory import Select, Order, Direction
        >>> query = Select(order=Order([('column1', Direction.ASC), ('column2', Direction.DESC)]))

        :param order: Ordering specification - either instance of Order, or collection of columns and directions.
        """
        super().__init__(*args, **kwargs)
        if order:
            self._order: Optional[Order] = order if isinstance(order, Order) else Order(order)
        else:
            self._order = None

    def order_by(self, column: OrderColumn, direction: Direction) -> Self:
        """
        Add column to be used for ordering. Can be called multiple times, columns will be ordered by order of calls.

        Example:

        >>> from sqlfactory import Select, Direction
        >>> query = Select().order_by('column1', Direction.ASC).order_by('column2', Direction.DESC)

        :param column: Column to use for ordering
        :param direction: Ordering direction
        """
        if self._order is None:
            self._order = Order()

        self._order.append((column, direction))
        return self

    # pylint: disable=invalid-name
    def ORDER_BY(self, column: OrderColumn, direction: Direction) -> Self:
        """Alias for `WithOrder.order_by()` to be more SQL-like with all capitals."""
        return self.order_by(column, direction)


# Ordering argument for class init (specify directly instance of Order, or collection of columns and directions).
OrderArg = Order | Collection[tuple[OrderColumn, Direction]]



"""DELETE statement builder"""

from typing import Any, Collection, TypeAlias

from sqlfactory.condition.base import ConditionBase
from sqlfactory.entities import Table
from sqlfactory.execute import ExecutableStatement
from sqlfactory.mixins.join import WithJoin
from sqlfactory.mixins.limit import Limit, WithLimit
from sqlfactory.mixins.order import OrderArg, WithOrder
from sqlfactory.mixins.where import WithWhere
from sqlfactory.select.join import Join


class Delete(
    ExecutableStatement,
    WithWhere["Delete"],
    WithOrder["Delete"],
    WithLimit["Delete"],
    WithJoin["Delete"],
):
    # pylint: disable=too-many-ancestors
    """
    `DELETE` statement

    Usage:

    >>> Delete("table", where=In("id", [1, 2, 3]))
    >>> "DELETE FROM `table` WHERE `id` IN (1,2,3)"

    More advanced DELETE with JOINs is also supported:

    >>> Delete("table", delete=["table"], join=[Join("table2", on=Eq("table.id", "table2.id"))], where=Eq("table2.value", 10))
    >>> "DELETE `table` FROM `table` JOIN `table2` ON `table`.`id` = `table2`.`id` WHERE `table2`.`value` = 10"
    """

    def __init__(
        self,
        table: Table | str,
        where: ConditionBase | None = None,
        order: OrderArg | None = None,
        limit: Limit | None = None,
        *,
        delete: Collection[Table | str] | None = None,
        join: Collection[Join] | None = None,
    ) -> None:
        """
        :param table: Table to delete from
        :param where: WHERE condition
        :param order: Ordering of matched rows, usefull when limiting number of deleted rows.
        :param limit: Limit number of deleted rows.
        """
        super().__init__(where=where, order=order, limit=limit, join=join)

        self.table = table if isinstance(table, Table) else Table(table)
        """Main table to delete from."""

        if delete is not None and (not isinstance(delete, Collection) or isinstance(delete, (str, bytes))):
            raise TypeError("delete argument must be a collection of tables to delete from")

        self.delete = [d if isinstance(d, Table) else Table(d) for d in delete] if delete is not None else []
        """When using join, specify tables to delete from to prevent deleting from joined tables."""

    def __str__(self) -> str:
        """Construct the DELETE statement."""
        q: list[str] = []

        if not self.delete:
            q.append(f"DELETE FROM {self.table!s}")
        else:
            q.append(f"DELETE {', '.join(str(t) for t in self.delete)} FROM {self.table!s}")

        if self._join:
            q.extend(map(str, self._join))

        if self._where:
            q.append("WHERE")
            q.append(str(self._where))

        if self._order:
            q.append(str(self._order))

        if self._limit:
            q.append(str(self._limit))

        return " ".join(q)

    @property
    def args(self) -> list[Any]:
        """DELETE statement arguments."""
        out = []

        if self._join:
            for join in self._join:
                out.extend(join.args)

        return (
            out
            + (self._where.args if self._where else [])
            + (self._order.args if self._order else [])
            + (self._limit.args if self._limit else [])
        )


DELETE: TypeAlias = Delete  # pylint: disable=invalid-name
"""
Alias for Delete statement to provide better SQL compatibility, as SQL is often written in all caps.
"""



"""Information functions (https://mariadb.com/kb/en/information-functions/)."""

from typing import Any

from sqlfactory.entities import Column
from sqlfactory.func.base import Function
from sqlfactory.statement import Raw, Statement


class Benchmark(Function):
    # pylint: disable=too-few-public-methods
    """Executes an expression repeatedly."""

    def __init__(self, count: int, expression: Statement) -> None:
        super().__init__("BENCHMARK", count, expression)


class BinlogGtidPos(Function):
    # pylint: disable=too-few-public-methods
    """Returns a string representation of the corresponding GTID position."""

    def __init__(self) -> None:
        super().__init__("BINLOG_GTID_POS")


class Charset(Function):
    # pylint: disable=too-few-public-methods
    """Returns the character set."""

    def __init__(self) -> None:
        super().__init__("CHARSET")


class Coercibility(Function):
    # pylint: disable=too-few-public-methods
    """Returns the collation coercibility value of the string expression."""

    def __init__(self, expression: str) -> None:
        super().__init__("COERCIBILITY", expression)


class Collation(Function):
    # pylint: disable=too-few-public-methods
    """Collation of the string argument"""

    def __init__(self, expression: str) -> None:
        super().__init__("COLLATION", expression)


class Collate(Raw):
    # pylint: disable=too-few-public-methods
    """String with collation"""

    def __init__(self, expression: str | Statement, collation: str) -> None:
        super().__init__(
            f"{str(expression) if isinstance(expression, Statement) else '%s'} COLLATE {collation}",
            *(
                expression.args
                if isinstance(expression, Statement)
                else [expression]
                if not isinstance(expression, Statement)
                else []
            ),
        )


class ConnectionId(Function):
    # pylint: disable=too-few-public-methods
    """Connection ID"""

    def __init__(self) -> None:
        super().__init__("CONNECTION_ID")


class CurrentRole(Function):
    # pylint: disable=too-few-public-methods
    """Current role name"""

    def __init__(self) -> None:
        super().__init__("CURRENT_ROLE")


class CurrentUser(Function):
    # pylint: disable=too-few-public-methods
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("CURRENT_USER")


class Database(Function):
    # pylint: disable=too-few-public-methods
    """Current default database"""

    def __init__(self) -> None:
        super().__init__("DATABASE")


class DecodeHistogram(Function):
    # pylint: disable=too-few-public-methods
    """Returns comma separated numerics corresponding to a probability distribution"""

    def __init__(self, hist_type: Any, histogram: Any) -> None:
        super().__init__("DECODE_HISTOGRAM", hist_type, histogram)


class Default(Function):
    # pylint: disable=too-few-public-methods
    """Returns the default value for a table column"""

    def __init__(self, column: Column) -> None:
        super().__init__("DEFAULT", column)


class FoundRows(Function):
    # pylint: disable=too-few-public-methods
    """Returns the number of (potentially) returned rows if there was no LIMIT involved."""

    def __init__(self) -> None:
        super().__init__("FOUND_ROWS")


class LastInsertId(Function):
    # pylint: disable=too-few-public-methods
    """Returns the value generated for an AUTO_INCREMENT column by the previous INSERT statement."""

    def __init__(self) -> None:
        super().__init__("LAST_INSERT_ID")


class LastValue(Function):
    # pylint: disable=too-few-public-methods
    """Evaluates expression and returns the last."""

    def __init__(self, expr: Statement, *exprs: Statement) -> None:
        super().__init__("LAST_VALUE", expr, *exprs)


class RowNumber(Function):
    # pylint: disable=too-few-public-methods
    """Returns the number of accepted rows so far."""

    def __init__(self) -> None:
        super().__init__("ROW_NUMBER")


class Schema(Function):
    # pylint: disable=too-few-public-methods
    """Current default schema"""

    def __init__(self) -> None:
        super().__init__("SCHEMA")


class SessionUser(Function):
    # pylint: disable=too-few-public-methods
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("SESSION_USER")


class SystemUser(Function):
    # pylint: disable=too-few-public-methods
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("SYSTEM_USER")


class User(Function):
    # pylint: disable=too-few-public-methods
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("USER")


class Version(Function):
    # pylint: disable=too-few-public-methods
    """Database version"""

    def __init__(self) -> None:
        super().__init__("VERSION")



"""VALUES() function for usage in ON DUPLICATE KEY UPDATE statements. MySQL / MariaDB specific."""

from sqlfactory.entities import Column, ColumnArg
from sqlfactory.func.base import Function


class Values(Function):
    # pylint: disable=too-few-public-methods  # Too few public methods, as everything is handled by super class.
    """
    `VALUES(<column>)` for usage in `INSERT INTO ... ON DUPLICATE KEY UPDATE column = VALUES(column)` statements.
    """

    def __init__(self, column: ColumnArg) -> None:
        """
        :param column: Column to be used in `VALUES(<column>)` function.
        """
        if not isinstance(column, Column):
            column = Column(column)

        super().__init__("VALUES", column)



"""JOIN statements for SQL queries."""

from sqlfactory.mixins.join import CrossJoin, InnerJoin, Join, LeftJoin, LeftOuterJoin, RightJoin, RightOuterJoin

__all__ = [
    "CrossJoin",
    "InnerJoin",
    "Join",
    "LeftJoin",
    "LeftOuterJoin",
    "RightJoin",
    "RightOuterJoin",
]



"""Control flow functions"""

from typing import Any

from sqlfactory.func.base import Function
from sqlfactory.statement import Statement


class IfNull(Function):
    # pylint: disable=too-few-public-methods
    """If expr1 is not NULL, IFNULL() returns expr1; otherwise it returns expr2."""

    def __init__(self, expr1: Statement | Any, expr2: Statement | Any) -> None:
        super().__init__("IFNULL", expr1, expr2)


class NullIf(Function):
    # pylint: disable=too-few-public-methods
    """Returns NULL if expr1 = expr2 is true, otherwise returns expr1."""

    def __init__(self, expr1: Statement | Any, expr2: Statement | Any) -> None:
        super().__init__("NULLIF", expr1, expr2)


class If(Function):
    # pylint: disable=too-few-public-methods
    """If expr1 is TRUE (expr1 <> 0 and expr1 <> NULL) then IF() returns expr2; otherwise it returns expr3."""

    def __init__(self, expr: Statement | Any, if_true: Statement | Any, if_false: Statement | Any) -> None:
        super().__init__("IF", expr, if_true, if_false)



"""
SQL functions to be used in the statements. Includes functions known to MariaDB 10.11, additional functions can be defined
on demand or in your code.
"""

# pylint: disable=redefined-builtin
from sqlfactory.func import agg, base, control, datetime, enc, info, misc, numeric, str  # noqa: A004

__all__ = [
    "agg",
    "base",
    "control",
    "datetime",
    "enc",
    "info",
    "misc",
    "numeric",
    "str",
]



"""Mixins are not directly exported from this module, as it should not be needed to use them in user code.
Classes exported here, however, should be used as their instances are expected as mixin arguments."""

from sqlfactory.mixins import join, limit, order, where
from sqlfactory.mixins.limit import Limit
from sqlfactory.mixins.order import Direction, Order

__all__ = ["Direction", "Limit", "Order", "join", "limit", "order", "where"]



"""
SQLFactory is a Python library for building SQL queries in a programmatic way.

Quick start:

## SELECT ...

Main class of interest: `Select`

Quick example:

```python
from sqlfactory import Select, Eq

query = Select("id", "name", table="products", where=Eq("enabled", True))
```
## INSERT ...

Main class of interest: `Insert`

Quick example:

```python
from sqlfactory import Insert

query = Insert.into("products")("id", "name").values(
    (1, "Product 1"),
    (2, "Product 2"),
)
```

## UPDATE ...

Main class of interest: `Update`

Quick example:

```python
from sqlfactory import Update, Eq

query = Update("products", set={"enabled": False}, where=Eq("id", 1))
```

## DELETE ...

Main class of interest: `Delete`

Quick example:

```python
from sqlfactory import Delete, Eq

query = Delete("products", where=Eq("enabled", False))
```

"""

from sqlfactory import execute, func, mixins  # exported submodules
from sqlfactory.condition import (
    And,
    Between,
    Condition,
    ConditionBase,
    Eq,
    Equals,
    Ge,
    GreaterThan,
    GreaterThanOrEquals,
    Gt,
    In,
    Le,
    LessThan,
    LessThanOrEquals,
    Like,
    Lt,
    Ne,
    NotEquals,
    Or,
)
from sqlfactory.delete import DELETE, Delete
from sqlfactory.entities import Column, Table
from sqlfactory.insert import INSERT, Insert, Values
from sqlfactory.mixins import Direction, Limit, Order
from sqlfactory.select import (
    SELECT,
    Aliased,
    ColumnList,
    CrossJoin,
    InnerJoin,
    Join,
    LeftJoin,
    LeftOuterJoin,
    RightJoin,
    RightOuterJoin,
    Select,
    SelectColumn,
)
from sqlfactory.statement import Raw, Statement
from sqlfactory.update import UPDATE, Update

__all__ = [
    "DELETE",
    "INSERT",
    "SELECT",
    "UPDATE",
    "Aliased",
    "And",
    "Between",
    "Column",
    "ColumnList",
    "Condition",
    "ConditionBase",
    "CrossJoin",
    "Delete",
    "Direction",
    "Eq",
    "Equals",
    "Ge",
    "GreaterThan",
    "GreaterThanOrEquals",
    "Gt",
    "In",
    "InnerJoin",
    "Insert",
    "Join",
    "Le",
    "LeftJoin",
    "LeftOuterJoin",
    "LessThan",
    "LessThanOrEquals",
    "Like",
    "Limit",
    "Lt",
    "Ne",
    "NotEquals",
    "Or",
    "Order",
    "Raw",
    "RightJoin",
    "RightOuterJoin",
    "Select",
    "SelectColumn",
    "Statement",
    "Table",
    "Update",
    "Values",
    "execute",
    "func",
    "mixins",
]



"""Classes representing database entities."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlfactory.statement import Raw, Statement

if TYPE_CHECKING:
    from sqlfactory.condition.simple import Eq, Ge, Gt, Le, Lt, Ne  # pragma: nocover


class Expression(Statement):
    """
    Expression as statement

    Represents expression that can be used in SQL statements. It provides basic methods for creating complex
    expressions by combining columns, functions and literals together to one arithmetic function.
    """

    def __gt__(self, other: Statement | Any) -> Gt:
        """Shorthand to produce conditional SQL statement <column> > <other>."""
        from sqlfactory.condition.simple import Gt  # pylint: disable=import-outside-toplevel

        return Gt(self, other)

    def __ge__(self, other: Statement | Any) -> Ge:
        """Shorthand to produce conditional SQL statement <column> >= <other>."""
        from sqlfactory.condition.simple import Ge  # pylint: disable=import-outside-toplevel

        return Ge(self, other)

    def __lt__(self, other: Statement | Any) -> Lt:
        """Shorthand to produce conditional SQL statement <column> < <other>."""
        from sqlfactory.condition.simple import Lt  # pylint: disable=import-outside-toplevel

        return Lt(self, other)

    def __le__(self, other: Statement | Any) -> Le:
        """Shorthand to produce conditional SQL statement <column> <= <other>."""
        from sqlfactory.condition.simple import Le  # pylint: disable=import-outside-toplevel

        return Le(self, other)

    def __eq__(self, other: Statement | Any) -> Eq:  # type: ignore[override]
        """Shorthand to produce conditional SQL statement <column> = <other>."""
        from sqlfactory.condition.simple import Eq  # pylint: disable=import-outside-toplevel

        return Eq(self, other)

    def __ne__(self, other: Statement | Any) -> Ne:  # type: ignore[override]
        """Shorthand to produce conditional SQL statement <column> != <other>."""
        from sqlfactory.condition.simple import Ne  # pylint: disable=import-outside-toplevel

        return Ne(self, other)

    def __add__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} + {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __sub__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} - {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __mul__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} * {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __truediv__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} / {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __mod__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} % {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __and__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} & {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __or__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} | {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __xor__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} ^ {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __lshift__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} << {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __rshift__(self, other: Statement | Any) -> Expression:
        return RawExpression(
            f"({self!s} >> {str(other) if isinstance(other, Statement) else '%s'})",
            *self.args,
            *other.args if isinstance(other, Statement) else [other] if not isinstance(other, Statement) else [],
        )

    def __neg__(self) -> Expression:
        return RawExpression(f"(~{self!s})", *self.args)


class RawExpression(Expression, Raw):
    """Expression as result of another expression."""


class Column(Expression):
    """
    Column (optionally with table and database) as statement.

    >>> from sqlfactory import Column
    >>> Column("column")
    >>> "`column`"

    >>> from sqlfactory import Column
    >>> Column("table.column")
    >>> "`table`.`column`"

    >>> from sqlfactory import Column
    >>> Column("database.table.column")
    >>> "`database`.`table`.`column`"

    To use column with `Select` statement with alias, you can use `Aliased` or `SelectColumn` classes instead.

    The class also provides shorthand for creating conditional SQL statements and arithmetic SQL statements. You can use Column
    instance to directly create condition using simple operators (`==`, `!=`, `<`, `<=`, `>`, `>=`) or arithmetic operations
    (`+`, `-`, `*`, `/`, `%`).

    >>> from sqlfactory import Column
    >>> Column("table.column") == 5
    >>> # Produces Eq(Column("table.column"), 5)

    >>> from sqlfactory import Column
    >>> Column("table.column") + 5
    >>> # Produces Expression("`table`.`column` + %s", 5)
    """

    def __init__(self, column: str) -> None:
        super().__init__()

        self._column = column.split(".")
        if len(self._column) > 3:
            raise ValueError("Invalid column name (contains more than <database>.<table>.<column>).")

    def __str__(self) -> str:
        return ".".join(f"`{x}`" if not x.startswith("`") else x for x in self._column)

    @property
    def column(self) -> str:
        """Returns column part of the column name."""
        return self._column[-1]

    @property
    def table(self) -> str | None:
        """Returns table part of the column name, if specified. If column specification does not contain table name,
        returns None."""
        try:
            return self._column[-2]
        except IndexError:
            return None

    @property
    def database(self) -> str | None:
        """Returns database part of the column name, if specified. If column specification does not contain database
        name, returns None."""
        try:
            return self._column[-3]
        except IndexError:
            return None

    def __hash__(self) -> int:
        return hash(str(self))

    @property
    def args(self) -> list[Any]:
        """Column does not have any arguments, so this always returns empty list."""
        return []


class Table(Statement):
    """
    Table (optionally with database) as statement

    >>> from sqlfactory import Table
    >>> Table("table")
    >>> "`table`"

    >>> from sqlfactory import Table
    >>> Table("database.table")
    >>> "`database`.`table`"

    To produce table alias, use `Aliased` class:

    >>> from sqlfactory import Table, Aliased
    >>> Aliased(Table("database.table"), alias="t1")
    >>> "`database`.`table` AS `t1`"

    By accessing Table's undefined attributes, instance of Column is returned. This allows to easily access columns of
    that table, and reference them in SQL statements.

    >>> from sqlfactory import Table, Select
    >>> t = Table("table")
    >>> Select(t.id, t.name, table=t)
    >>> "SELECT `table`.`id`, `table`.`name` FROM `table`"

    This allows creating python-like expressions that are converted to SQL automatically:

    >>> from sqlfactory import Table, Select
    >>> from sqlfactory.func.agg import Count
    >>> t = Table("database.table")
    >>> Select(Count(t.id), table=t, where=t.id > 5)
    >>> "SELECT COUNT(`database`.`table`.`id`) FROM `database`.`table` WHERE `database`.`table`.`id` > %s"
    """

    def __init__(self, table: str) -> None:
        """
        :param table: Table name (optionally with database in form `<database>.<table>`). If database is not specified,
            it is assumed that table is in default database and SQL constructed will not contain any database name.
        """
        super().__init__()
        self._table = table.split(".")

        if len(self._table) > 2:
            raise ValueError("Invalid table name (contains more than <database>.<table>).")

    def __str__(self) -> str:
        return ".".join(f"`{x}`" if not x.startswith("`") else x for x in self._table)

    @property
    def table(self) -> str | None:
        """Returns table part of the table name"""
        return self._table[-1]

    @property
    def database(self) -> str | None:
        """Returns database part of the table name, if specified. If table specification does not contain database,
        returns None."""
        try:
            return self._table[-2]
        except IndexError:
            return None

    def __getattr__(self, name: str) -> Column:
        """Returns column of that table."""
        return Column(f"{'.'.join(self._table)}.{name}")

    @property
    def args(self) -> list[Any]:
        """Table does not have any arguments, this always returns empty list."""
        return []


# Alias for column that can be passed as instance of column or as string.
ColumnArg = Column | str



"""Aggregate functions."""

from typing import Literal

from sqlfactory.entities import Column, ColumnArg
from sqlfactory.func.base import Function
from sqlfactory.statement import Raw, Statement


class AggregateFunction(Function):
    # pylint: disable=too-few-public-methods
    """Base class for aggregate functions"""

    def __init__(self, agg: str, column: ColumnArg | Statement):
        super().__init__(agg, Column(column) if isinstance(column, str) else column)


class Avg(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """AVG(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("AVG", column)


class BitAnd(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """BIT_AND(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("BIT_AND", column)


class BitOr(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """BIT_OR(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("BIT_OR", column)


class BitXor(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """BIT_XOR(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("BIT_XOR", column)


class Count(Function):
    # pylint: disable=too-few-public-methods
    """
    - COUNT(<column>)
    - COUNT(DISTINCT <column>)
    """

    def __init__(self, column: ColumnArg | Literal["*"], *, distinct: bool = False):
        if isinstance(column, str) and column == "*":
            column_stmt: Statement = Raw("*")
        elif isinstance(column, str):
            column_stmt = Column(column)
        else:
            column_stmt = column

        if distinct:
            super().__init__(
                "COUNT", Raw(f"DISTINCT {column_stmt!s}", *column_stmt.args if isinstance(column_stmt, Statement) else [])
            )
        else:
            super().__init__("COUNT", column_stmt)


class Max(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """MAX(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("MAX", column)


class Min(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """MIN(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("MIN", column)


class Std(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """STD(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("STD", column)


class Sum(AggregateFunction):
    # pylint: disable=too-few-public-methods
    """SUM(<column>)"""

    def __init__(self, column: ColumnArg | Statement):
        super().__init__("SUM", column)



"""UPDATE statement builder."""

from __future__ import annotations

from typing import Any, Optional, Self, TypeAlias

from sqlfactory.condition.base import ConditionBase
from sqlfactory.entities import Column, ColumnArg, Table
from sqlfactory.execute import ConditionalExecutableStatement
from sqlfactory.mixins.limit import Limit, WithLimit
from sqlfactory.mixins.where import WithWhere
from sqlfactory.statement import Statement


class UpdateColumn(Statement):
    """
    Represents one field that should be updated.
    """

    def __init__(self, column: ColumnArg, value: Statement | Any) -> None:
        self._column = column if isinstance(column, Column) else Column(column)
        self._value = value

    def __str__(self) -> str:
        return f"{self._column!s} = {str(self._value) if isinstance(self._value, Statement) else '%s'}"

    def __hash__(self) -> int:
        return hash(self._column)

    def __eq__(self, other: UpdateColumn | Any) -> bool:
        if not isinstance(other, UpdateColumn):
            return False

        return str(self._column) == str(other._column)

    @property
    def args(self) -> list[Any]:
        """
        Return arguments for the update statement.
        """
        if isinstance(self._value, Statement):
            return self._value.args

        return [self._value]


class Update(ConditionalExecutableStatement, WithWhere["Update"], WithLimit["Update"]):
    # pylint: disable=too-many-ancestors  # This is intentional, as this class is a combination of multiple mixins.
    """
    Builds `UPDATE` statement SQL query.

    This is conditional SQL statement, so you can check whether it should be executed (would update any columns)
    by calling bool() on it. Also, if you are using execute() method of the statement, the execution won't be
    performed if bool() returns False.

    Examples:

    >>> from sqlfactory import Update, Eq
    >>> Update("table").set("column1", 1).where(Eq("column2", 2))

    >>> from sqlfactory import Update, Table
    >>> t = Table("table")
    >>> Update(t).set(t.column1, 1).where(t.column2 == 2)
    """

    def __init__(
        self, table: Table | str, *fields: UpdateColumn, where: Optional[ConditionBase] = None, limit: Optional[Limit] = None
    ) -> None:
        """
        :param table: Table to update.
        :param fields: List of UpdateColumn instances containing columns to be updated. This is not very pleasant way
            to create the statement, use set() method instead.
        :param where: WHERE condition
        :param limit: Limit number of updated rows
        """
        super().__init__(where=where, limit=limit)

        self.table = table if isinstance(table, Table) else Table(table)
        """Table that should be updated."""

        self.fields: list[UpdateColumn] = list(fields)
        """Fields that should be updated."""

    def __str__(self) -> str:
        """
        Return the UPDATE statement with %s placeholders for arguments.
        """
        if not self.fields:
            raise AttributeError("At least one column must be updated.")

        query = [f"UPDATE {self.table!s}", f"SET {', '.join(map(str, self.fields))}"]

        if self._where:
            query.append("WHERE")
            query.append(str(self._where))

        if self._limit:
            query.append(str(self._limit))

        return " ".join(query)

    @property
    def args(self) -> list[Any]:
        """
        Return all arguments used in the UPDATE statement.
        """
        out = []

        for field in self.fields:
            out.extend(field.args)

        if self._where is not None:
            out.extend(self._where.args)

        if self._limit is not None:
            out.extend(self._limit.args)

        return out

    def __bool__(self) -> bool:
        """
        Return True if there are any fields to be updated and the statement should be executed.
        """
        return bool(self.fields)

    def append(self, field: UpdateColumn) -> Self:
        """
        Append new UpdateField to this UPDATE statement. Can be used when set() method is not sufficient.
        """
        if field in self.fields:
            raise AttributeError(f"Field '{field}' is already in the list of fields to be updated.")

        self.fields.append(field)
        return self

    def set(self, field: ColumnArg, value: Statement | Any) -> Self:
        """
        Syntactical sugar for creating simple SET UpdateFields.
        :param field: Field name (without quotes).
        :param value: Value to set the field to (will be escaped).
        """
        return self.append(UpdateColumn(field, value))

    def SET(self, field: str, value: Any) -> Self:
        # pylint: disable=invalid-name
        """Alias for `Update.set()` for better SQL compatibility (SQL is often written in all caps)."""
        return self.set(field, value)


UPDATE: TypeAlias = Update  # pylint: disable=invalid-name
"""
Alias for Update statement to provide better SQL compatibility, as SQL is often written in all caps.
"""
